use {
    crate::{error::Error, B32Alphabet},
    std::io::{self, ErrorKind, Read, Write},
};

pub struct Encoder<R: Read, W: Write> {
    reader: R,
    writer: W,
    alphabet: B32Alphabet,
    wrap: Option<usize>,
}

impl<R: Read, W: Write> Encoder<R, W> {
    pub fn new(reader: R, writer: W, alphabet: Option<B32Alphabet>, wrap: Option<usize>) -> Self {
        Self {
            reader,
            writer,
            alphabet: alphabet.unwrap_or_default(),
            wrap,
        }
    }

    pub fn output(self) -> W {
        self.writer
    }

    pub fn encode(&mut self) -> Result<(), Error> {
        let mut total: usize = 0;
        loop {
            let mut buf = [0; 5];
            let mut num: u64 = 0;
            let n_bytes;
            loop {
                n_bytes = match self.reader.read(&mut buf) {
                    Ok(n) => n,
                    Err(e) if e.kind() == ErrorKind::Interrupted => continue,
                    Err(e) => return Err(e.into()),
                };
                break;
            }
            if n_bytes == 0 {
                break;
            }
            for (idx, c) in buf.iter().enumerate() {
                num <<= 8;
                if idx < n_bytes {
                    num |= *c as u64;
                }
            }
            let mut buf = [b'\0'; 8];
            for i in (0..8).rev() {
                let b = (num & 0x1f) as u8;
                buf[i] = self.alphabet.items[b as usize];
                num >>= 5;
            }
            let mut outlen = n_bytes * 8 / 5;
            let rem = n_bytes * 8 % 5;
            if rem != 0 {
                outlen += 1;
            }
            buf[0..outlen].iter().try_for_each(|c| {
                self.writer.write_all(&[*c])?;
                total += 1;
                if let Some(wrap) = self.wrap {
                    if total % wrap == 0 {
                        writeln!(self.writer)?;
                    }
                }
                Ok::<(), io::Error>(())
            })?;
            if let Some(pad) = self.alphabet.pad {
                if rem != 0 {
                    for _c in 0..8 - outlen {
                        self.writer.write_all(&[pad])?;
                    }
                }
            }
            if n_bytes < 5 {
                break;
            }
        }
        Ok(())
    }
}

#[cfg(test)]
mod test {
    use {
        super::*,
        std::{
            fs::{self, File},
            io::BufReader,
        },
    };

    #[test]
    fn hello() {
        let reader = "Hello, World".as_bytes();
        let mut encoder = Encoder::new(reader, vec![], None, None);
        encoder.encode().unwrap();
        assert_eq!(encoder.writer, b"JBSWY3DPFQQFO33SNRSA====");
        encoder = Encoder::new("Hello, World!".as_bytes(), vec![], None, None);
        encoder.encode().unwrap();
        assert_eq!(encoder.writer, b"JBSWY3DPFQQFO33SNRSCC===");
        encoder = Encoder::new("Hello, World!\n".as_bytes(), vec![], None, None);
        encoder.encode().unwrap();
        assert_eq!(encoder.writer, b"JBSWY3DPFQQFO33SNRSCCCQ=");
    }

    #[test]
    fn encode_from_file() {
        let infile = File::open("src/testdata/lorem.txt").unwrap();
        let reader = BufReader::new(infile);
        let outfile = fs::read_to_string("src/testdata/lorem_b32.txt").unwrap();
        let mut encoder = Encoder::new(reader, vec![], None, Some(76));
        encoder.encode().unwrap();
        assert_eq!(encoder.writer, outfile.as_bytes());
    }
}
